defmodule ClickHouse.Interface.HTTP do
  @moduledoc """
  An interface to interact with a ClickHouse server via HTTP.
  """

  @behaviour ClickHouse.Interface

  use GenServer

  require Logger

  alias ClickHouse.{ConnectionError, Query, Result}

  @pool_opts_schema %{
    timeout: [type: :integer, required: true, default: 150_000],
    max_connections: [type: :integer, required: true, default: 50]
  }

  @opts_schema %{
    client: [type: :atom, required: true, default: :default],
    urls: [type: {:list, :binary}, required: true, default: ["http://localhost:8123"]],
    connect_retry: [type: :integer, required: true, default: 3_000],
    pool_opts: [type: {:keyword, @pool_opts_schema}, required: true, default: []]
  }

  @execute_opts_schema %{
    checkout_timeout: [type: :timeout, required: false, default: 8_000],
    connect_timeout: [type: :timeout, required: false, default: 8_000],
    recv_timeout: [type: :timeout, required: false, default: 15_000],
    buffer_size: [type: :integer, required: false],
    wait_end_of_query: [type: :integer, required: false],
    session_id: [type: :binary, required: false],
    session_timeout: [type: :integer, required: false],
    database: [type: :binary, required: false],
    settings: [type: :map, required: false],
    content_encoding: [type: :atom, inclusion: [:gzip, :deflate], required: false],
    accept_encoding: [type: :atom, inclusion: [:gzip, :deflate], required: false],
    decompress: [type: :boolean, required: false]
  }

  @ping_statement "SELECT 1"
  @accept_header "Accept-Encoding"
  @content_header "Content-Encoding"
  @format_header "X-ClickHouse-Format"
  @error_header "X-ClickHouse-Exception-Code"
  @error_regex ~r/(?<=displayText\(\) = )(.*?)(?=: )/
  @error_types %{
    "DB::Exception" => ClickHouse.DatabaseError,
    "DB::ParsingException" => ClickHouse.ParsingError
  }

  ################################
  # ClickHouse.Interface Callbacks
  ################################

  @doc """
  Starts the HTTP interface server.
  """
  @impl ClickHouse.Interface
  @spec start_link(keyword()) :: GenServer.on_start()
  def start_link(opts) do
    opts = KeywordValidator.validate!(opts, @opts_schema)
    GenServer.start_link(__MODULE__, opts)
  end

  @doc """
  Executes a query via HTTP.
  """
  @impl ClickHouse.Interface
  @spec execute(ClickHouse.Query.t()) ::
          {:error, ClickHouse.Interface.error()} | {:ok, ClickHouse.Result.t()}
  def execute(query) do
    opts = KeywordValidator.validate!(query.opts, @execute_opts_schema)

    with {:ok, {pool, urls}} <- get_config(query.client) do
      statement = Query.get_statement(query)
      do_execute(query.client, pool, urls, statement, opts)
    end
  end

  ################################
  # GenServer Callbacks
  ################################

  @impl GenServer
  def init(opts) do
    state = init_state(opts)
    init_config(state)
    init_pool(state)
    {:ok, state, {:continue, :connect}}
  end

  @impl GenServer
  def handle_continue(:connect, state) do
    connect(state)
    {:noreply, state}
  end

  @impl GenServer
  def handle_info(:retry_connect, state) do
    {:noreply, state, {:continue, :connect}}
  end

  ################################
  # Private API
  ################################

  defp init_state(opts) do
    client = Keyword.get(opts, :client)

    %{
      client: client,
      table_name: table_name(client),
      pool_name: pool_name(client),
      pool_opts: Keyword.get(opts, :pool_opts),
      urls: Keyword.get(opts, :urls),
      connect_retry: Keyword.get(opts, :connect_retry)
    }
  end

  defp init_config(state) do
    config = {state.pool_name, state.urls}
    :ets.new(state.table_name, [:named_table, read_concurrency: true])
    :ets.insert(state.table_name, {:config, config})
  end

  defp get_config(client) do
    [{:config, config}] = :ets.lookup(table_name(client), :config)
    {:ok, config}
  rescue
    ArgumentError ->
      raise ArgumentError, "No client interface available"
  end

  defp init_pool(state) do
    :hackney_pool.start_pool(state.pool_name, state.pool_opts)
  end

  defp table_name(client) do
    :"#{__MODULE__}.#{client}"
  end

  defp pool_name(client) do
    :"#{__MODULE__}.#{client}.http_pool"
  end

  defp connect(state) do
    case do_execute(state.client, state.pool_name, state.urls, @ping_statement, []) do
      {:ok, _} ->
        :ok

      {:error, error} ->
        Logger.error("[ClickHouse] Connection error: #{inspect(error)}")
        Process.send_after(self(), :retry_connect, state.connect_retry)
        :error
    end
  end

  defp do_execute(client, pool, [url], statement, opts) do
    do_execute(client, pool, url, statement, opts)
  end

  defp do_execute(client, pool, urls, statement, opts) when is_list(urls) do
    url = Enum.random(urls)
    do_execute(client, pool, url, statement, opts)
  end

  defp do_execute(client, pool, url, statement, opts) when is_binary(url) do
    url = build_url(url, opts)
    headers = build_headers(opts)
    statement = maybe_compress(statement, opts)
    http_opts = build_http_opts(pool, opts)

    case :hackney.request(:post, url, headers, statement, http_opts) do
      {:ok, 200, headers, body} -> build_result(client, headers, body, opts)
      error -> handle_error(error)
    end
  end

  defp build_url(url, opts) do
    settings = Keyword.get(opts, :settings, %{})

    params =
      opts
      |> maybe_compression_param()
      |> Keyword.take([
        :buffer_size,
        :wait_end_of_query,
        :session_id,
        :session_timeout,
        :database,
        :enable_http_compression
      ])
      |> Enum.into(%{})
      |> Map.merge(settings)

    case URI.encode_query(params) do
      "" -> url
      params -> "#{url}?#{params}"
    end
  end

  defp maybe_compression_param(opts) do
    if Keyword.has_key?(opts, :accept_encoding) do
      Keyword.put(opts, :enable_http_compression, 1)
    else
      opts
    end
  end

  defp build_headers(opts) do
    []
    |> maybe_accept_encoding(opts)
    |> maybe_content_encoding(opts)
  end

  defp maybe_accept_encoding(headers, opts) do
    if accept_encoding = Keyword.get(opts, :accept_encoding) do
      [{@accept_header, to_string(accept_encoding)} | headers]
    else
      headers
    end
  end

  defp maybe_content_encoding(headers, opts) do
    if content_encoding = Keyword.get(opts, :content_encoding) do
      [{@content_header, to_string(content_encoding)} | headers]
    else
      headers
    end
  end

  defp maybe_compress(statement, opts) do
    if compression = Keyword.get(opts, :content_encoding) do
      compress(compression, statement)
    else
      statement
    end
  end

  defp compress(:gzip, statement), do: :zlib.gzip(statement)
  defp compress(:deflate, statement), do: :zlib.compress(statement)

  defp build_http_opts(pool, opts) do
    opts
    |> Keyword.take([:checkout_timeout, :connect_timeout, :recv_timeout])
    |> Keyword.merge(pool: pool, with_body: true)
  end

  defp build_result(client, headers, body, opts) do
    format = find_header(headers, @format_header)
    encoding = find_header(headers, @content_header)
    {body, compressed} = maybe_decompress(encoding, body, opts)
    meta = build_result_meta(headers)
    result = Result.new(client, body, format, meta, compressed)

    {:ok, result}
  end

  defp build_result_meta(headers) do
    %{
      query_id: find_header(headers, "X-ClickHouse-Query-Id"),
      server_display_name: find_header(headers, "X-ClickHouse-Server-Display-Name"),
      timezone: find_header(headers, "X-ClickHouse-Timezone"),
      summary: find_header(headers, "X-ClickHouse-Summary")
    }
  end

  defp handle_error({:error, reason}) do
    {:error, %ConnectionError{message: to_string(reason)}}
  end

  defp handle_error({:ok, _code, headers, body}) do
    encoding = find_header(headers, @content_header)
    {body, _} = maybe_decompress(encoding, body, decompress: true)
    error_type = fetch_error_type(body)
    error_code = find_header(headers, @error_header)
    error_message = fetch_error_message(body)
    error = struct(error_type, code: error_code, message: error_message)

    {:error, error}
  end

  defp fetch_error_type(body) do
    case Regex.run(@error_regex, body, capture: :first) do
      [error] -> Map.fetch!(@error_types, error)
    end
  end

  defp fetch_error_message(body) do
    case String.split(body, "e.displayText() = ") do
      [_, message] -> message
    end
  end

  defp find_header([], _), do: nil

  defp find_header([{key, val} | _headers], header) when key == header do
    val
  end

  defp find_header([_ | headers], header) do
    find_header(headers, header)
  end

  defp maybe_decompress(nil, body, _opts), do: {body, false}

  defp maybe_decompress(compression, body, opts) do
    if Keyword.get(opts, :decompress) do
      {decompress(compression, body), false}
    else
      {body, true}
    end
  end

  defp decompress("gzip", body), do: :zlib.gunzip(body)
  defp decompress("deflate", body), do: :zlib.uncompress(body)
end
